class HorcruxFileWriter():
    """Wraps a horcrux file for writing.

        |------------------------------headers-----------------------------|--blocks--|
     => |HXL1|hash128|nonce|n|k|pos|blocksize|numblocks|share[0]|share[1]|&| .|.|.|.  |

        n,k,pos,blocksize,numblocks,share[0] are all 32 bit unsigned integers (big endian).
        hash128 and share[1] are 128 bit unsigned integers (big endian).
    """


    def __init__(self, filename):
        self.file = open(filename, 'wb')
    
    def write_headers(self, sourcehash: int, 
                            nonce: bytes, 
                            n: int, 
                            k: int, 
                            pos: int, 
                            blocksize: int, 
                            numblocks:int, 
                            share: tuple):

        """ Writes headers to file """

        self.file.write(b"HXL1")
        self.file.write(sourcehash.to_bytes(128//8, 'big'))
        self.file.write(nonce)

        for x in [n, k, pos, blocksize, numblocks, share[0]]:
            self.file.write(x.to_bytes(32//8, byteorder='big'))

        self.file.write(share[1].to_bytes(128//8, 'big'))

        self.file.write(b'&')

    def write_block(self, block: bytes):
        self.file.write(block)

    def close(self):
        self.file.close()


class HorcruxFileReader():
    """Wraps a horcrux file for reading.

        |------------------------------headers-----------------------------|--blocks--|
     => |HXL1|hash128|nonce|n|k|pos|blocksize|numblocks|share[0]|share[1]|&| .|.|.|.  |

        n,k,pos,blocksize,numblocks,share[0] are all 32 bit unsigned integers (big endian).
        hash128 and share[1] are 128 bit unsigned integers (big endian).
    """

    def __init__(self, filename):
        self.file = open(filename, 'rb')
        name = self.file.read(4)
        if name != b"HXL1": raise ValueError(f"File {filename} is not a horcrux")
        self.hashval = int.from_bytes(self.file.read(16), byteorder="big")

        def read_int():
            return int.from_bytes(self.file.read(4), byteorder="big")

        self.nonce = self.file.read(8)
        self.n =         read_int()
        self.k =         read_int()
        self.pos =       read_int()
        self.blocksize = read_int()
        self.numblocks = read_int()
        self.share =    (read_int(), int.from_bytes(self.file.read(128//8), "big"))
        if self.file.read(1) != b"&": 
            raise ValueError("Header not terminated. File may be corrupted")

        self.headerend = self.file.tell()
    
    def read_block_at(self, offset):
        self.file.seek(self.headerend + offset * self.blocksize, 0)
        return self.file.read(self.blocksize)


"""
    Implement's Shamir's Secret Sharing.
"""
import secrets 

MAX_PRIME = 2 ** 127 - 1  

def eval_polynomial_at(poly: tuple, x: int):
    """Evaluates a tuple polynomial at a point"""
    accum = 0
    for coeff in reversed(poly):
        accum *= x
        accum += coeff
        accum %= MAX_PRIME
    return accum

def generate_shares(secret: int, n: int, k: int) -> list:
    """Generates n shares such that k are required to recreate `secret`"""
    poly = secret, *tuple(secrets.randbelow(MAX_PRIME) for _ in range(k-1))

    points = [(x, eval_polynomial_at(poly, x)) for x in range(1, n+1)]

    return points
    
def extended_gcd(a, b):
    x = 0
    last_x = 1
    y = 1
    last_y = 0
    while b != 0:
        quot = a // b
        a, b = b, a % b
        x, last_x = last_x - quot * x, x
        y, last_y = last_y - quot * y, y
    return last_x, last_y

def divide_mod(num, den, p):
    """Compute num / den modulo prime p"""
    inv, _ = extended_gcd(den, p)
    return num * inv


def lagrange_interpolate(x, x_s, y_s, p):
    """
    Find the y-value for the given x, given n (x, y) points;
    k points will define a polynomial of up to kth order.
    """
    k = len(x_s)
    assert k == len(set(x_s)), "points must be distinct"
    def PI(vals):  # upper-case PI -- product of inputs
        accum = 1
        for v in vals:
            accum *= v
        return accum
    nums = []  # avoid inexact division
    dens = []
    for i in range(k):
        others = list(x_s)
        cur = others.pop(i)
        nums.append(PI(x - o for o in others))
        dens.append(PI(cur - o for o in others))
    den = PI(dens)
    num = sum([divide_mod(nums[i] * den * y_s[i] % p, dens[i], p)
               for i in range(k)])
    return (divide_mod(num, den, p) + p) % p


def find_secret(shares: list) -> int:
    x_s, y_s = zip(*shares)
    return lagrange_interpolate(0, x_s, y_s, MAX_PRIME)


import unittest

class TestShamir(unittest.TestCase):
    def test_shamir(self):
        import random
        from tqdm import tqdm

        tests = 50

        print(f"Running {tests} test cases...")
        for i in tqdm(range(tests)):
            secret = secrets.randbelow(MAX_PRIME)
            n = random.randrange(3,40)
            k = random.randrange(2,n)

            shares = generate_shares(secret, n, k)


            for i in range(5):
                random.shuffle(shares)
                recovered = find_secret(shares[:k])
                self.assertEqual(secret, recovered)

if __name__ == '__main__':
    unittest.main()


from horcruxes.cruxio import HorcruxFileReader
from Crypto.Cipher import AES

class HorcruxReverseManager():
    def __init__(self, filenames, dest):
        self.files = []
        self.dest = open(dest, 'wb')
        for file in filenames:
            self.files.append(HorcruxFileReader(file))

        if len(set(file.hashval for file in self.files)) > 1:
            raise ValueError("Horcruxes belong to different files.")
        
        if len(set(file.nonce for file in self.files)) > 1:
            raise ValueError("Horcruxes do not have same nonce.")

        if self.files[0].k > len(self.files):
            raise ValueError(f"Not enough horcruxes. Required {self.files[0].k}, got {len(self.files)}")

        self.nonce = self.files[0].nonce
        self.n=self.files[0].n
        self.k=self.files[0].k
        self.filenos = [file.pos for file in self.files]
        self.numblocks = self.files[0].numblocks
        self.blocksize = self.files[0].blocksize
        self.secret = self.get_secret()
        self.aes = AES.new(self.secret.to_bytes(16, 'big'), AES.MODE_CTR, nonce=self.nonce)

    def get_secret(self):
        import horcruxes.shamir as shamir
        shares = [file.share for file in self.files]

        return shamir.find_secret(shares)

    def decrypt(self):
        from tqdm import tqdm

        for bno in tqdm(range(self.numblocks), "Decrypting"):
            self.decrypt_block(bno)


    def decrypt_block(self, bno):
        fileno = bno*(self.n-self.k+1) % self.n
        offset = bno*(self.n-self.k+1)
        
        for _ in range(self.n-self.k+1):
            if fileno in self.filenos: break
            fileno += 1
            offset += 1
            fileno %= self.n
        else:
            raise AssertionError("Something's gone horribly wrong. A reqired block was not found in any file.")

        offset = offset//self.n
        
        block = self.files[self.filenos.index(fileno)].read_block_at(offset)
        decrypted = self.aes.decrypt(block)
        self.dest.write(decrypted)


if __name__ == "__main__":
    import os
    files = [ 'output/' + x for x in  os.listdir('./output/')[:5]] 
    dest = 'uncrux.mkv'

    hrm = HorcruxReverseManager(files, dest)
    hrm.decrypt()




import click
import secrets

@click.group()
def cli():
    pass


@cli.command()
@click.option('-n', type=int, required=True, help='Number of horcruxes to create')
@click.option('-k', type=int, required=True, help='Number of horcruxes to recover original')
@click.option('-o', default='./output', type=click.Path(exists=False), help='Destination directory')
@click.option('--block-size', type=int, help="Size of block to operate on. Larger values are faster, but may result in horcruxes of different sizes. Defaults to 1/10th of file size")
@click.argument('file', type=click.Path(exists=True, readable=True))
def split(file, n: int, k: int, o, block_size):
    """Splits a file into horcruxes"""
    import os
    os.makedirs(o, exist_ok=True)

    if n > k: 
        print("WARN: n > k. Data will be irrecoverable. What a waste of CPU cycles.")

    from horcruxes.cruxcreator import HorcruxCreateManager
    hcm = HorcruxCreateManager(file, n, k, block_size, o)
    hcm.write_headers()
    hcm.write()

    print("Operation successful.")

@cli.command()
@click.option('-o', type=click.Path(exists=False), required=True, help='Destination directory')
@click.argument('files-or-dir', type=click.Path(exists=True, readable=True), nargs=-1)
def bind(files_or_dir, o):
    """Binds horcruxes back into the original. files-or-dir may be a list of files or a directory containing a list of horcrux files"""
    import os
    if os.path.dirname(o): os.makedirs(os.path.dirname(o), exist_ok=True)

    from horcruxes.cruxreverser import HorcruxReverseManager
    files = files_or_dir
    if len(files_or_dir) == 1 and os.path.isdir(files_or_dir[0]):
        files = [os.path.join(files_or_dir[0], file) for file in os.listdir(files_or_dir[0])]

    try:
        hrm = HorcruxReverseManager(files, o)
        hrm.decrypt()
    except ValueError as e:
        import sys
        print(f"Invalid arguments: {e}", file=sys.stderr)
    else:
        print("Operation successful.")

if __name__ == '__main__':
    cli()


from setuptools import setup

setup(name='horcruxes',
      version='2.7',
      url='https://github.com/X-yl/horcruxes',
      description='A command line tool to split a file into horcruxes, where the original can be recreated with at least n of the horcruxes',
      author='X-yl',
      long_description= open('README.md').read(),
      license='GNU GPLv3',
      long_description_content_type='text/markdown',
      entry_points={
        "console_scripts": [
            "horcruxes = horcruxes.crux:cli",
        ]
      },
      install_requires=[
            "click>=7.1.2",
            "mmh3>=2.5.1",
            "pycryptodome>=3.9.8",
            "tqdm>=4.48.2",
      ],
      python_requires=">=3.8",
      packages=['horcruxes'])



from horcruxes.cruxio import HorcruxFileWriter
from Crypto.Cipher import AES

class HorcruxCreateManager():

    def __init__(self, source, n, k, blocksize, outdir):
        import os
        import secrets
        import horcruxes.shamir as shamir
        from math import ceil
        self.blocksize = blocksize or os.stat(source).st_size//10
        self.numblocks = ceil(os.stat(source).st_size/self.blocksize)
        self.source = open(source, 'rb')
        self.n = n
        self.k = k
        self.secret = secrets.randbelow(shamir.MAX_PRIME)
        self.shares = shamir.generate_shares(self.secret, n, k)
        self.files = []
        self.aes = AES.new(self.secret.to_bytes(16, byteorder='big'), mode=AES.MODE_CTR)
        self.nonce = self.aes.nonce
        for i in range(n):
           self.files.append(HorcruxFileWriter(os.path.join(outdir, f"{source}_{i+1}_of_{n}.hcx")))

    def write_headers(self):
        file: HorcruxFile = None
        srchash = self.hash_source()
        for i,file in enumerate(self.files):
            file.write_headers(srchash, self.nonce, self.n, self.k, i, self.blocksize, self.numblocks, self.shares[i])
        

    def hash_source(self):
        """Obtains a murmurhash3 hash of source file"""
        import mmh3
        val = 0
        while chunk:=self.source.read(self.blocksize):
            val = mmh3.hash128(chunk, val)
        
        self.source.seek(0)
        return val

    def write_block(self, bno):
        fileno = bno * (self.n-self.k+1) % self.n
        block = self.source.read(self.blocksize)

        encrypted = self.aes.encrypt(block)
        for _ in range(self.n-self.k+1):
            self.files[fileno].write_block(encrypted)
            fileno += 1
            fileno %= self.n

    def write(self):
        from tqdm import tqdm
        for i in tqdm(range(self.numblocks), desc="Encrypting"):
            self.write_block(i)
